@@ -1,5 +1,5 @@
 # This file is part of the python-chess library.
-# Copyright (C) 2012-2020 Niklas Fiekas <niklas.fiekas@backscattering.de>
+# Copyright (C) 2012-2021 Niklas Fiekas <niklas.fiekas@backscattering.de>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -14,19 +14,28 @@
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
 
+from __future__ import annotations
+
 import abc
 import enum
 import itertools
 import logging
 import re
-import weakref
 import typing
 
 import chess
 import chess.engine
 import chess.svg
 
-from typing import Callable, Dict, Generic, Iterable, Iterator, List, Mapping, MutableMapping, Set, TextIO, Tuple, Type, TypeVar, Optional, Union
+from typing import Any, Callable, Dict, Generic, Iterable, Iterator, List, Mapping, MutableMapping, Set, TextIO, Tuple, Type, TypeVar, Optional, Union
+from chess import Color, Square
+
+try:
+    from typing import Literal
+    _TrueLiteral = Literal[True]
+except ImportError:
+    # Before Python 3.8.
+    _TrueLiteral = bool  # type: ignore
 
 
 LOGGER = logging.getLogger(__name__)
@@ -94,6 +103,8 @@ MOVETEXT_REGEX = re.compile(r"""
         |[PNBRQK]?@[a-h][1-8]
         |--
         |Z0
+        |0000
+        |@@@@
         |O-O(?:-O)?
         |0-0(?:-0)?
     )
@@ -109,13 +120,16 @@ MOVETEXT_REGEX = re.compile(r"""
 SKIP_MOVETEXT_REGEX = re.compile(r""";|\{|\}""")
 
 
-CLOCK_REGEX = re.compile(r"""\[%clk\s(\d+):(\d+):(\d+)\]""")
+CLOCK_REGEX = re.compile(r"""\[%clk\s(\d+):(\d+):(\d+(?:\.\d*)?)\]""")
+EMT_REGEX = re.compile(r"""\[%emt\s(\d+):(\d+):(\d+(?:\.\d*)?)\]""")
 
 EVAL_REGEX = re.compile(r"""
     \[%eval\s(?:
         \#([+-]?\d+)
         |([+-]?(?:\d{0,10}\.\d{1,2}|\d{1,10}\.?))
-    )\]
+    )(?:
+        ,(\d+)
+    )?\]
     """, re.VERBOSE)
 
 ARROWS_REGEX = re.compile(r"""
@@ -139,8 +153,7 @@ ResultT = TypeVar("ResultT", covariant=True)
 
 
 class _AcceptFrame:
-    def __init__(self, node: "GameNode", *, is_variation: bool = False, sidelines: bool = True):
-        assert node.parent is not None, "cannot visit dangling GameNode"
+    def __init__(self, node: ChildNode, *, is_variation: bool = False, sidelines: bool = True):
         self.state = "pre"
         self.node = node
         self.is_variation = is_variation
@@ -148,25 +161,40 @@ class _AcceptFrame:
         self.in_variation = False
 
 
-class GameNode:
-    def __init__(self) -> None:
-        self.parent: Optional[GameNode] = None
-        self.variations: List[GameNode] = []
+class GameNode(abc.ABC):
+    parent: Optional[GameNode]
+    """The parent node or ``None`` if this is the root node of the game."""
 
-        self.move: Optional[chess.Move] = None
-        self.nags: Set[int] = set()
+    move: Optional[chess.Move]
+    """
+    The move leading to this node or ``None`` if this is the root node of the
+    game.
+    """
+
+    variations: List[ChildNode]
+    """A list of child nodes."""
+
+    comment: str
+    """
+    A comment that goes behind the move leading to this node. Comments
+    that occur before any moves are assigned to the root node.
+    """
+
+    starting_comment: str
+    nags: Set[int]
+
+    def __init__(self, *, comment: str = "") -> None:
+        self.parent = None
+        self.move = None
+        self.variations = []
+        self.comment = comment
+
+        # Deprecated: These should be properties of ChildNode, but need to
+        # remain here for backwards compatibility.
         self.starting_comment = ""
-        self.comment = ""
-
-        self.board_cached: Optional[weakref.ref[chess.Board]] = None
-
-    @classmethod
-    def dangling_node(cls) -> "GameNode":
-        return GameNode()
-
-    def _board(self) -> chess.Board:
-        assert False, "cannot get board of dangling GameNode"
+        self.nags = set()
 
+    @abc.abstractmethod
     def board(self) -> chess.Board:
         """
         Gets a board with the position of the node.
@@ -176,70 +204,37 @@ class GameNode:
 
         It's a copy, so modifying the board will not alter the game.
         """
-        if self.board_cached is not None:
-            board_cached = self.board_cached()
-            if board_cached is not None:
-                return board_cached.copy()
 
-        stack = []
-        node = self
-        while True:
-            if node.parent is None or node.move is None:
-                board = node._board()
-                break
-            elif node.board_cached is not None:
-                board_cached = node.board_cached()
-                if board_cached is not None:
-                    board = board_cached.copy()
-                    break
-            stack.append(node.move)
-            node = node.parent
-
-        while stack:
-            board.push(stack.pop())
-
-        if self.parent is not None:
-            self.board_cached = weakref.ref(board)
-
-        return board
-
-    def _move(self) -> chess.Move:
-        assert self.move is not None, "cannot get move of dangling GameNode"
-        return self.move
-
-    def san(self) -> str:
+    @abc.abstractmethod
+    def ply(self) -> int:
         """
-        Gets the standard algebraic notation of the move leading to this node.
-        See :func:`chess.Board.san()`.
+        Returns the number of half-moves up to this node, as indicated by
+        fullmove number and turn of the position.
+        See :func:`chess.Board.ply()`.
 
-        Do not call this on the root node.
+        Usually this is equal to the number of parent nodes, but it may be
+        more if the game was started from a custom position.
         """
-        assert self.parent is not None and self.move is not None, "cannot get san of dangling GameNode"
-        return self.parent.board().san(self.move)
 
-    def uci(self, *, chess960: Optional[bool] = None) -> str:
+    def turn(self) -> Color:
         """
-        Gets the UCI notation of the move leading to this node.
-        See :func:`chess.Board.uci()`.
-
-        Do not call this on the root node.
+        Gets the color to move at this node. See :data:`chess.Board.turn`.
         """
-        assert self.parent is not None and self.move is not None, "cannot get uci of dangling GameNode"
-        return self.parent.board().uci(self.move, chess960=chess960)
+        return self.ply() % 2 == 0
 
-    def root(self) -> "GameNode":
+    def root(self) -> GameNode:
         node = self
         while node.parent:
             node = node.parent
         return node
 
-    def game(self) -> "Game":
+    def game(self) -> Game:
         """Gets the root node, i.e., the game."""
         root = self.root()
         assert isinstance(root, Game), "GameNode not rooted in Game"
         return root
 
-    def end(self) -> "GameNode":
+    def end(self) -> GameNode:
         """Follows the main variation to the end and returns the last node."""
         node = self
 
@@ -290,7 +285,7 @@ class GameNode:
 
         return not self.parent.variations or self.parent.variations[0] == self
 
-    def __getitem__(self, move: Union[int, chess.Move]) -> "GameNode":
+    def __getitem__(self, move: Union[int, chess.Move, GameNode]) -> ChildNode:
         try:
             return self.variations[move]  # type: ignore
         except TypeError:
@@ -300,71 +295,78 @@ class GameNode:
 
         raise KeyError(move)
 
-    def variation(self, move: Union[int, chess.Move]) -> "GameNode":
+    def __contains__(self, move: Union[int, chess.Move, GameNode]) -> bool:
+        try:
+            self[move]
+        except KeyError:
+            return False
+        else:
+            return True
+
+    def variation(self, move: Union[int, chess.Move, GameNode]) -> ChildNode:
         """
         Gets a child node by either the move or the variation index.
         """
         return self[move]
 
-    def has_variation(self, move: chess.Move) -> bool:
-        """Checks if the given *move* appears as a variation."""
-        return move in (variation.move for variation in self.variations)
+    def has_variation(self, move: Union[int, chess.Move, GameNode]) -> bool:
+        """Checks if this node has the given variation."""
+        return move in self
 
-    def promote_to_main(self, move: chess.Move) -> None:
+    def promote_to_main(self, move: Union[int, chess.Move, GameNode]) -> None:
         """Promotes the given *move* to the main variation."""
         variation = self[move]
         self.variations.remove(variation)
         self.variations.insert(0, variation)
 
-    def promote(self, move: chess.Move) -> None:
+    def promote(self, move: Union[int, chess.Move, GameNode]) -> None:
         """Moves a variation one up in the list of variations."""
         variation = self[move]
         i = self.variations.index(variation)
         if i > 0:
             self.variations[i - 1], self.variations[i] = self.variations[i], self.variations[i - 1]
 
-    def demote(self, move: chess.Move) -> None:
+    def demote(self, move: Union[int, chess.Move, GameNode]) -> None:
         """Moves a variation one down in the list of variations."""
         variation = self[move]
         i = self.variations.index(variation)
         if i < len(self.variations) - 1:
             self.variations[i + 1], self.variations[i] = self.variations[i], self.variations[i + 1]
 
-    def remove_variation(self, move: chess.Move) -> None:
+    def remove_variation(self, move: Union[int, chess.Move, GameNode]) -> None:
         """Removes a variation."""
         self.variations.remove(self.variation(move))
 
-    def add_variation(self, move: chess.Move, *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> "GameNode":
+    def add_variation(self, move: chess.Move, *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> ChildNode:
         """Creates a child node with the given attributes."""
-        node = type(self).dangling_node()
-        node.move = move
-        node.nags = set(nags)
-        node.comment = comment
-        node.starting_comment = starting_comment
+        # Instanciate ChildNode only in this method.
+        return ChildNode(self, move, comment=comment, starting_comment=starting_comment, nags=nags)
 
-        node.parent = self
-        self.variations.append(node)
-        return node
-
-    def add_main_variation(self, move: chess.Move, *, comment: str = "") -> "GameNode":
+    def add_main_variation(self, move: chess.Move, *, comment: str = "", nags: Iterable[int] = []) -> ChildNode:
         """
         Creates a child node with the given attributes and promotes it to the
         main variation.
         """
-        node = self.add_variation(move, comment=comment)
-        self.variations.remove(node)
-        self.variations.insert(0, node)
+        node = self.add_variation(move, comment=comment, nags=nags)
+        self.variations.insert(0, self.variations.pop())
         return node
 
-    def mainline(self) -> "Mainline[GameNode]":
-        """Returns an iterator over the mainline starting after this node."""
+    def next(self) -> Optional[ChildNode]:
+        """
+        Returns the first node of the mainline after this node, or ``None`` if
+        this node does not have any children.
+        """
+        return self.variations[0] if self.variations else None
+
+    def mainline(self) -> Mainline[ChildNode]:
+        """Returns an iterable over the mainline starting after this node."""
         return Mainline(self, lambda node: node)
 
-    def mainline_moves(self) -> "Mainline[chess.Move]":
-        """Returns an iterator over the main moves after this node."""
-        return Mainline(self, lambda node: node._move())
+    def mainline_moves(self) -> Mainline[chess.Move]:
+        """Returns an iterable over the main moves after this node."""
+        return Mainline(self, lambda node: node.move)
 
-    def add_line(self, moves: Iterable[chess.Move], *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> "GameNode":
+    def add_line(self, moves: Iterable[chess.Move], *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> GameNode:
         """
         Creates a sequence of child nodes for the given list of moves.
         Adds *comment* and *nags* to the last node of the line and returns it.
@@ -394,24 +396,43 @@ class GameNode:
         match = EVAL_REGEX.search(self.comment)
         if not match:
             return None
+
+        turn = self.turn()
+
         if match.group(1):
             mate = int(match.group(1))
-            return chess.engine.PovScore(chess.engine.Mate(mate), chess.WHITE) if mate else None
+            score: chess.engine.Score = chess.engine.Mate(mate)
+            if mate == 0:
+                # Resolve this ambiguity in the specification in favor of
+                # standard chess: The player to move after mate is the player
+                # who has been mated.
+                return chess.engine.PovScore(score, turn)
         else:
-            return chess.engine.PovScore(chess.engine.Cp(int(float(match.group(2)) * 100)), chess.WHITE)
+            score = chess.engine.Cp(int(float(match.group(2)) * 100))
 
-    def set_eval(self, score: Optional[chess.engine.PovScore]) -> None:
+        return chess.engine.PovScore(score if turn else -score, turn)
+
+    def eval_depth(self) -> Optional[int]:
+        """
+        Parses the first valid ``[%eval ...]`` annotation in the comment of
+        this node and returns the corresponding depth, if any.
+        """
+        match = EVAL_REGEX.search(self.comment)
+        return int(match.group(3)) if match and match.group(3) else None
+
+    def set_eval(self, score: Optional[chess.engine.PovScore], depth: Optional[int] = None) -> None:
         """
         Replaces the first valid ``[%eval ...]`` annotation in the comment of
         this node or adds a new one.
         """
         eval = ""
         if score is not None:
+            depth_suffix = "" if depth is None else f",{max(depth, 0):d}"
             cp = score.white().score()
             if cp is not None:
-                eval = f"[%eval {float(cp) / 100:.2f}]"
+                eval = f"[%eval {float(cp) / 100:.2f}{depth_suffix}]"
             elif score.white().mate():
-                eval = f"[%eval #{score.white().mate()}]"
+                eval = f"[%eval #{score.white().mate()}{depth_suffix}]"
 
         self.comment, found = EVAL_REGEX.subn(eval, self.comment, count=1)
 
@@ -430,50 +451,25 @@ class GameNode:
         arrows = []
         for match in ARROWS_REGEX.finditer(self.comment):
             for group in match.group(1).split(","):
-                color = "G"
-                if group[0] == "R":
-                    color = "red"
-                elif group[0] == "Y":
-                    color = "yellow"
-                elif group[0] == "B":
-                    color = "blue"
-
-                tail = chess.SQUARE_NAMES.index(group[1:3])
-                if len(group) == 5:
-                    head = chess.SQUARE_NAMES.index(group[3:5])
-                else:
-                    head = tail
-
-                arrows.append(chess.svg.Arrow(tail, head, color=color))
+                arrows.append(chess.svg.Arrow.from_pgn(group))
 
         return arrows
 
-    def set_arrows(self, arrows: Iterable[Union[chess.svg.Arrow, Tuple[chess.Square, chess.Square]]]) -> None:
+    def set_arrows(self, arrows: Iterable[Union[chess.svg.Arrow, Tuple[Square, Square]]]) -> None:
         """
         Replaces all valid ``[%csl ...]`` and ``[%cal ...]`` annotations in
         the comment of this node or adds new ones.
         """
-        csl = []
-        cal = []
+        csl: List[str] = []
+        cal: List[str] = []
 
         for arrow in arrows:
             try:
-                tail, head, color = arrow.tail, arrow.head, arrow.color  # type: ignore
-                if color == "red":
-                    color = "R"
-                elif color == "yellow":
-                    color = "Y"
-                elif color == "blue":
-                    color = "B"
-                else:
-                    color = "G"
-            except AttributeError:
                 tail, head = arrow  # type: ignore
-                color = "G"
-            if tail == head:
-                csl.append(f"{color}{chess.SQUARE_NAMES[tail]}")
-            else:
-                cal.append(f"{color}{chess.SQUARE_NAMES[tail]}{chess.SQUARE_NAMES[head]}")
+                arrow = chess.svg.Arrow(tail, head)
+            except TypeError:
+                pass
+            (csl if arrow.tail == arrow.head else cal).append(arrow.pgn())  # type: ignore
 
         self.comment = ARROWS_REGEX.sub("", self.comment).strip()
 
@@ -497,7 +493,7 @@ class GameNode:
         match = CLOCK_REGEX.search(self.comment)
         if match is None:
             return None
-        return int(match.group(1)) * 3600 + int(match.group(2)) * 60 + int(match.group(3))
+        return int(match.group(1)) * 3600 + int(match.group(2)) * 60 + float(match.group(3))
 
     def set_clock(self, seconds: Optional[float]) -> None:
         """
@@ -506,11 +502,12 @@ class GameNode:
         """
         clk = ""
         if seconds is not None:
-            seconds = max(0, round(seconds))
-            hours = seconds // 3600
-            minutes = seconds % 3600 // 60
+            seconds = max(0, seconds)
+            hours = int(seconds // 3600)
+            minutes = int(seconds % 3600 // 60)
             seconds = seconds % 3600 % 60
-            clk = f"[%clk {hours:d}:{minutes:02d}:{seconds:02d}]"
+            seconds_part = f"{seconds:06.3f}".rstrip("0").rstrip(".")
+            clk = f"[%clk {hours:d}:{minutes:02d}:{seconds_part}]"
 
         self.comment, found = CLOCK_REGEX.subn(clk, self.comment, count=1)
 
@@ -519,68 +516,48 @@ class GameNode:
                 self.comment += " "
             self.comment += clk
 
-    def _accept_node(self, parent_board: chess.Board, visitor: "BaseVisitor[ResultT]") -> None:
-        assert self.move is not None, "cannot visit dangling GameNode"
+    def emt(self) -> Optional[float]:
+        """
+        Parses the first valid ``[%emt ...]`` annotation in the comment of
+        this node, if any.
 
-        if self.starting_comment:
-            visitor.visit_comment(self.starting_comment)
+        Returns the player's elapsed move time use for the comment of this
+        move, in seconds.
+        """
+        match = EMT_REGEX.search(self.comment)
+        if match is None:
+            return None
+        return int(match.group(1)) * 3600 + int(match.group(2)) * 60 + float(match.group(3))
 
-        visitor.visit_move(parent_board, self.move)
+    def set_emt(self, seconds: Optional[float]) -> None:
+        """
+        Replaces the first valid ``[%emt ...]`` annotation in the comment of
+        this node or adds a new one.
+        """
+        emt = ""
+        if seconds is not None:
+            seconds = max(0, seconds)
+            hours = int(seconds // 3600)
+            minutes = int(seconds % 3600 // 60)
+            seconds = seconds % 3600 % 60
+            seconds_part = f"{seconds:06.3f}".rstrip("0").rstrip(".")
+            emt = f"[%emt {hours:d}:{minutes:02d}:{seconds_part}]"
 
-        parent_board.push(self.move)
-        visitor.visit_board(parent_board)
-        parent_board.pop()
+        self.comment, found = EMT_REGEX.subn(emt, self.comment, count=1)
 
-        for nag in sorted(self.nags):
-            visitor.visit_nag(nag)
+        if not found and emt:
+            if self.comment and not self.comment.endswith(" "):
+                self.comment += " "
+            self.comment += emt
 
-        if self.comment:
-            visitor.visit_comment(self.comment)
-
-    def _accept(self, parent_board: chess.Board, visitor: "BaseVisitor[ResultT]", *, sidelines: bool = True) -> None:
-        stack = [_AcceptFrame(self, sidelines=sidelines)]
-
-        while stack:
-            top = stack[-1]
-
-            if top.in_variation:
-                top.in_variation = False
-                visitor.end_variation()
-
-            if top.state == "pre":
-                top.node._accept_node(parent_board, visitor)
-                top.state = "variations"
-            elif top.state == "variations":
-                try:
-                    variation = next(top.variations)
-                except StopIteration:
-                    if top.node.variations:
-                        assert top.node.move is not None, "root node cannot be a variation"
-                        parent_board.push(top.node.move)
-                        stack.append(_AcceptFrame(top.node.variations[0], sidelines=True))
-                        top.state = "post"
-                    else:
-                        top.state = "end"
-                else:
-                    if visitor.begin_variation() is not SKIP:
-                        stack.append(_AcceptFrame(variation, sidelines=False, is_variation=True))
-                    top.in_variation = True
-            elif top.state == "post":
-                parent_board.pop()
-                top.state = "end"
-            else:
-                stack.pop()
-
-    def accept(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
+    @abc.abstractmethod
+    def accept(self, visitor: BaseVisitor[ResultT]) -> ResultT:
         """
         Traverses game nodes in PGN order using the given *visitor*. Starts with
         the move leading to this node. Returns the *visitor* result.
         """
-        assert self.parent is not None, "cannot visit dangling GameNode"
-        self._accept(self.parent.board(), visitor, sidelines=False)
-        return visitor.result()
 
-    def accept_subgame(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
+    def accept_subgame(self, visitor: BaseVisitor[ResultT]) -> ResultT:
         """
         Traverses headers and game nodes in PGN order, as if the game was
         starting after this node. Returns the *visitor* result.
@@ -614,10 +591,141 @@ class GameNode:
     def __str__(self) -> str:
         return self.accept(StringExporter(columns=None))
 
-    def __repr__(self) -> str:
-        if self.parent is None:
-            return f"<{type(self).__name__} at {id(self):#x} (dangling: {self.move})>"
 
+class ChildNode(GameNode):
+    """
+    A child node of a game, with the move leading to it.
+    Extends :class:`~chess.pgn.GameNode`.
+    """
+
+    parent: GameNode
+    """The parent node."""
+
+    move: chess.Move
+    """The move leading to this node."""
+
+    starting_comment: str
+    """
+    A comment for the start of a variation. Only nodes that
+    actually start a variation (:func:`~chess.pgn.GameNode.starts_variation()`
+    checks this) can have a starting comment. The root node can not have
+    a starting comment.
+    """
+
+    nags: Set[int]
+    """
+    A set of NAGs as integers. NAGs always go behind a move, so the root
+    node of the game will never have NAGs.
+    """
+
+    def __init__(self, parent: GameNode, move: chess.Move, *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> None:
+        super().__init__(comment=comment)
+        self.parent = parent
+        self.move = move
+        self.parent.variations.append(self)
+
+        self.nags.update(nags)
+        self.starting_comment = starting_comment
+
+    def board(self) -> chess.Board:
+        stack: List[chess.Move] = []
+        node: GameNode = self
+
+        while node.move is not None and node.parent is not None:
+            stack.append(node.move)
+            node = node.parent
+
+        board = node.game().board()
+
+        while stack:
+            board.push(stack.pop())
+
+        return board
+
+    def ply(self) -> int:
+        ply = 0
+        node: GameNode = self
+        while node.parent is not None:
+            ply += 1
+            node = node.parent
+        return node.game().ply() + ply
+
+    def san(self) -> str:
+        """
+        Gets the standard algebraic notation of the move leading to this node.
+        See :func:`chess.Board.san()`.
+
+        Do not call this on the root node.
+        """
+        return self.parent.board().san(self.move)
+
+    def uci(self, *, chess960: Optional[bool] = None) -> str:
+        """
+        Gets the UCI notation of the move leading to this node.
+        See :func:`chess.Board.uci()`.
+
+        Do not call this on the root node.
+        """
+        return self.parent.board().uci(self.move, chess960=chess960)
+
+    def end(self) -> ChildNode:
+        """Follows the main variation to the end and returns the last node."""
+        return typing.cast(ChildNode, super().end())
+
+    def _accept_node(self, parent_board: chess.Board, visitor: BaseVisitor[ResultT]) -> None:
+        if self.starting_comment:
+            visitor.visit_comment(self.starting_comment)
+
+        visitor.visit_move(parent_board, self.move)
+
+        parent_board.push(self.move)
+        visitor.visit_board(parent_board)
+        parent_board.pop()
+
+        for nag in sorted(self.nags):
+            visitor.visit_nag(nag)
+
+        if self.comment:
+            visitor.visit_comment(self.comment)
+
+    def _accept(self, parent_board: chess.Board, visitor: BaseVisitor[ResultT], *, sidelines: bool = True) -> None:
+        stack = [_AcceptFrame(self, sidelines=sidelines)]
+
+        while stack:
+            top = stack[-1]
+
+            if top.in_variation:
+                top.in_variation = False
+                visitor.end_variation()
+
+            if top.state == "pre":
+                top.node._accept_node(parent_board, visitor)
+                top.state = "variations"
+            elif top.state == "variations":
+                try:
+                    variation = next(top.variations)
+                except StopIteration:
+                    if top.node.variations:
+                        parent_board.push(top.node.move)
+                        stack.append(_AcceptFrame(top.node.variations[0], sidelines=True))
+                        top.state = "post"
+                    else:
+                        top.state = "end"
+                else:
+                    if visitor.begin_variation() is not SKIP:
+                        stack.append(_AcceptFrame(variation, sidelines=False, is_variation=True))
+                    top.in_variation = True
+            elif top.state == "post":
+                parent_board.pop()
+                top.state = "end"
+            else:
+                stack.pop()
+
+    def accept(self, visitor: BaseVisitor[ResultT]) -> ResultT:
+        self._accept(self.parent.board(), visitor, sidelines=False)
+        return visitor.result()
+
+    def __repr__(self) -> str:
         try:
             parent_board = self.parent.board()
         except ValueError:
@@ -628,7 +736,7 @@ class GameNode:
                 id(self),
                 parent_board.fullmove_number,
                 "." if parent_board.turn == chess.WHITE else "...",
-                self.san())
+                parent_board.san(self.move))
 
 
 GameT = TypeVar("GameT", bound="Game")
@@ -636,18 +744,44 @@ GameT = TypeVar("GameT", bound="Game")
 class Game(GameNode):
     """
     The root node of a game with extra information such as headers and the
-    starting position. Also has all the other properties and methods of
-    :class:`~chess.pgn.GameNode`.
+    starting position. Extends :class:`~chess.pgn.GameNode`.
+    """
+
+    headers: Headers
+    """
+    A mapping of headers. By default, the following 7 headers are provided
+    (Seven Tag Roster):
+
+    >>> import chess.pgn
+    >>>
+    >>> game = chess.pgn.Game()
+    >>> game.headers
+    Headers(Event='?', Site='?', Date='????.??.??', Round='?', White='?', Black='?', Result='*')
+    """
+
+    errors: List[Exception]
+    """
+    A list of errors (such as illegal or ambiguous moves) encountered while
+    parsing the game.
     """
 
     def __init__(self, headers: Optional[Union[Mapping[str, str], Iterable[Tuple[str, str]]]] = None) -> None:
         super().__init__()
         self.headers = Headers(headers)
-        self.errors: List[Exception] = []
+        self.errors = []
 
-    def _board(self) -> chess.Board:
+    def board(self) -> chess.Board:
         return self.headers.board()
 
+    # TODO: Consider naming.
+    def _interactive_viewer(self) -> Any:
+        from chess._interactive import InteractiveViewer
+        return InteractiveViewer(self)  # type: ignore
+
+    def ply(self) -> int:
+        # Optimization: Parse FEN only for custom starting positions.
+        return self.board().ply() if "FEN" in self.headers else 0
+
     def setup(self, board: Union[chess.Board, str]) -> None:
         """
         Sets up a specific starting position. This sets (or resets) the
@@ -676,7 +810,7 @@ class Game(GameNode):
         else:
             self.headers.pop("Variant", None)
 
-    def accept(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
+    def accept(self, visitor: BaseVisitor[ResultT]) -> ResultT:
         """
         Traverses the game in PGN order using the given *visitor*. Returns
         the *visitor* result.
@@ -720,7 +854,7 @@ class Game(GameNode):
         return cls(headers={})
 
     @classmethod
-    def builder(cls: Type[GameT]) -> "GameBuilder[GameT]":
+    def builder(cls: Type[GameT]) -> GameBuilder[GameT]:
         return GameBuilder(Game=cls)
 
     def __repr__(self) -> str:
@@ -826,14 +960,14 @@ class Headers(MutableMapping[str, str]):
             ", ".join("{}={!r}".format(key, value) for key, value in self.items()))
 
     @classmethod
-    def builder(cls: Type[HeadersT]) -> "HeadersBuilder[HeadersT]":
+    def builder(cls: Type[HeadersT]) -> HeadersBuilder[HeadersT]:
         return HeadersBuilder(Headers=cls)
 
 
 MainlineMapT = TypeVar("MainlineMapT")
 
 class Mainline(Generic[MainlineMapT]):
-    def __init__(self, start: GameNode, f: Callable[[GameNode], MainlineMapT]) -> None:
+    def __init__(self, start: GameNode, f: Callable[[ChildNode], MainlineMapT]) -> None:
         self.start = start
         self.f = f
 
@@ -846,16 +980,19 @@ class Mainline(Generic[MainlineMapT]):
             node = node.variations[0]
             yield self.f(node)
 
-    def __reversed__(self) -> "ReverseMainline[MainlineMapT]":
-        return ReverseMainline(self.start, self.f)
+    def __reversed__(self) -> Iterator[MainlineMapT]:
+        node = self.start.end()
+        while node.parent and node != self.start:
+            yield self.f(typing.cast(ChildNode, node))
+            node = node.parent
 
-    def accept(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
+    def accept(self, visitor: BaseVisitor[ResultT]) -> ResultT:
         node = self.start
         board = self.start.board()
         while node.variations:
             node = node.variations[0]
             node._accept_node(board, visitor)
-            board.push(node._move())
+            board.push(node.move)
         return visitor.result()
 
     def __str__(self) -> str:
@@ -865,36 +1002,6 @@ class Mainline(Generic[MainlineMapT]):
         return f"<Mainline at {id(self):#x} ({self.accept(StringExporter(columns=None, comments=False))})>"
 
 
-class ReverseMainline(Generic[MainlineMapT]):
-    def __init__(self, stop: GameNode, f: Callable[[GameNode], MainlineMapT]) -> None:
-        self.stop = stop
-        self.f = f
-
-        self.length = 0
-        node = stop
-        while node.variations:
-            node = node.variations[0]
-            self.length += 1
-        self.end = node
-
-    def __len__(self) -> int:
-        return self.length
-
-    def __iter__(self) -> Iterator[MainlineMapT]:
-        node = self.end
-        while node.parent and node != self.stop:
-            yield self.f(node)
-            node = node.parent
-
-    def __reversed__(self) -> Mainline[MainlineMapT]:
-        return Mainline(self.stop, self.f)
-
-    def __repr__(self) -> str:
-        return "<ReverseMainline at {:#x} ({})>".format(
-            id(self),
-            " ".join(ReverseMainline(self.stop, lambda node: node._move().uci())))
-
-
 class BaseVisitor(abc.ABC, Generic[ResultT]):
     """
     Base class for visitors.
@@ -928,13 +1035,13 @@ class BaseVisitor(abc.ABC, Generic[ResultT]):
 
         You can override the default implementation to work around specific
         quirks of your input format.
-        """
-        # Replace zeros with correct castling notation.
-        if san == "0-0":
-            san = "O-O"
-        elif san == "0-0-0":
-            san = "O-O-O"
 
+        .. deprecated:: 1.1
+            This method is very limited, because it is only called on moves
+            that the parser recognizes in the first place. Instead of adding
+            workarounds here, please report common quirks so that
+            they can be handled for everyone.
+        """
         return board.parse_san(san)
 
     def visit_move(self, board: chess.Board, move: chess.Move) -> None:
@@ -998,10 +1105,10 @@ class GameBuilder(BaseVisitor[GameT]):
     """
 
     @typing.overload
-    def __init__(self: "GameBuilder[Game]") -> None: ...
+    def __init__(self: GameBuilder[Game]) -> None: ...
     @typing.overload
-    def __init__(self: "GameBuilder[GameT]", *, Game: Type[GameT]) -> None: ...
-    def __init__(self, *, Game = Game) -> None:
+    def __init__(self: GameBuilder[GameT], *, Game: Type[GameT]) -> None: ...
+    def __init__(self, *, Game: Any = Game) -> None:
         self.Game = Game
 
     def begin_game(self) -> None:
@@ -1074,7 +1181,7 @@ class GameBuilder(BaseVisitor[GameT]):
         >>> import chess.pgn
         >>>
         >>> class MyGameBuilder(chess.pgn.GameBuilder):
-        >>>     def handle_error(self, error):
+        >>>     def handle_error(self, error: Exception) -> None:
         >>>         pass  # Ignore error
         >>>
         >>> pgn = open("data/pgn/kasparov-deep-blue-1997.pgn")
@@ -1095,10 +1202,10 @@ class HeadersBuilder(BaseVisitor[HeadersT]):
     """Collects headers into a dictionary."""
 
     @typing.overload
-    def __init__(self: "HeadersBuilder[Headers]") -> None: ...
+    def __init__(self: HeadersBuilder[Headers]) -> None: ...
     @typing.overload
-    def __init__(self: "HeadersBuilder[HeadersT]", *, Headers: Type[Headers]) -> None: ...
-    def __init__(self, *, Headers = Headers) -> None:
+    def __init__(self: HeadersBuilder[HeadersT], *, Headers: Type[Headers]) -> None: ...
+    def __init__(self, *, Headers: Any = Headers) -> None:
         self.Headers = Headers
 
     def begin_headers(self) -> HeadersT:
@@ -1139,7 +1246,7 @@ class BoardBuilder(BaseVisitor[chess.Board]):
         return self.board
 
 
-class SkipVisitor(BaseVisitor["typing.Literal[True]"]):
+class SkipVisitor(BaseVisitor[_TrueLiteral]):
     """Skips a game."""
 
     def begin_game(self) -> SkipType:
@@ -1151,7 +1258,7 @@ class SkipVisitor(BaseVisitor["typing.Literal[True]"]):
     def begin_variation(self) -> SkipType:
         return SKIP
 
-    def result(self) -> "typing.Literal[True]":
+    def result(self) -> _TrueLiteral:
         return True
 
 
@@ -1320,7 +1427,7 @@ class FileExporter(StringExporterMixin, BaseVisitor[int]):
 def read_game(handle: TextIO) -> Optional[Game]: ...
 @typing.overload
 def read_game(handle: TextIO, *, Visitor: Callable[[], BaseVisitor[ResultT]]) -> Optional[ResultT]: ...
-def read_game(handle: TextIO, *, Visitor = GameBuilder):
+def read_game(handle: TextIO, *, Visitor: Any = GameBuilder) -> Any:
     """
     Reads a game from a file opened in text mode.
 
@@ -1344,10 +1451,10 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
 
     By using text mode, the parser does not need to handle encodings. It is the
     caller's responsibility to open the file with the correct encoding.
-    PGN files are usually ASCII or UTF-8 encoded. So, the following should
-    cover most relevant cases (ASCII, UTF-8, UTF-8 with BOM).
+    PGN files are usually ASCII or UTF-8 encoded, sometimes with BOM (which
+    this parser automatically ignores).
 
-    >>> pgn = open("data/pgn/kasparov-deep-blue-1997.pgn", encoding="utf-8-sig")
+    >>> pgn = open("data/pgn/kasparov-deep-blue-1997.pgn", encoding="utf-8")
 
     Use :class:`~io.StringIO` to parse games from a string.
 
@@ -1366,7 +1473,7 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
     The parser is relatively forgiving when it comes to errors. It skips over
     tokens it can not parse. By default, any exceptions are logged and
     collected in :data:`Game.errors <chess.pgn.Game.errors>`. This behavior can
-    be :func:`overriden <chess.pgn.GameBuilder.handle_error>`.
+    be :func:`overridden <chess.pgn.GameBuilder.handle_error>`.
 
     Returns the parsed game or ``None`` if the end of file is reached.
     """
@@ -1440,12 +1547,14 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
         # Initial position.
         fen = headers.get("FEN", VariantBoard.starting_fen)
         try:
-            board_stack = [VariantBoard(fen, chess960=headers.is_chess960())]
+            board = VariantBoard(fen, chess960=headers.is_chess960())
         except ValueError as error:
             visitor.handle_error(error)
             skipping_game = True
         else:
-            visitor.visit_board(board_stack[0])
+            board.chess960 = board.chess960 or board.has_chess960_castling_rights()
+            board_stack = [board]
+            visitor.visit_board(board)
 
     # Fast path: Skip entire game.
     if skipping_game:
@@ -1523,9 +1632,12 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
                         board.pop()
                         board_stack.append(board)
             elif token == ")":
-                if skip_variation_depth:
+                if skip_variation_depth == 1:
+                    skip_variation_depth = 0
+                    visitor.end_variation()
+                elif skip_variation_depth:
                     skip_variation_depth -= 1
-                if len(board_stack) > 1:
+                elif len(board_stack) > 1:
                     visitor.end_variation()
                     board_stack.pop()
             elif skip_variation_depth:
@@ -1570,7 +1682,8 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
 
 def read_headers(handle: TextIO) -> Optional[Headers]:
     """
-    Reads game headers from a PGN file opened in text mode.
+    Reads game headers from a PGN file opened in text mode. Skips the rest of
+    the game.
 
     Since actually parsing many games from a big file is relatively expensive,
     this is a better way to look only for specific games and then seek and
