@@ -1,5 +1,5 @@
 # This file is part of the python-chess library.
-# Copyright (C) 2012-2020 Niklas Fiekas <niklas.fiekas@backscattering.de>
+# Copyright (C) 2012-2021 Niklas Fiekas <niklas.fiekas@backscattering.de>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -14,6 +14,8 @@
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
 
+from __future__ import annotations
+
 import collections
 import math
 import mmap
@@ -378,9 +380,6 @@ def is_tablename(name: str, *, one_king: bool = True, piece_count: Optional[int]
         (not normalized or normalize_tablename(name) == name) and
         (not one_king or (name != "KvK" and name.startswith("K") and "vK" in name)))
 
-def is_table_name(name: str) -> bool:  # TODO: Remove before 1.0
-    return is_tablename(name, one_king=False)
-
 
 def tablenames(*, one_king: bool = True, piece_count: int = 6) -> Iterator[str]:
     first = "K" if one_king else "P"
@@ -1536,8 +1535,8 @@ class Tablebase:
         """
         Adds tables from a directory.
 
-        By default all available tables with the correct file names
-        (e.g. WDL files like ``KQvKN.rtbw`` and DTZ files like ``KRBvK.rtbz``)
+        By default, all available tables with the correct file names
+        (e.g., WDL files like ``KQvKN.rtbw`` and DTZ files like ``KRBvK.rtbz``)
         are added.
 
         The relevant files are lazily opened when the tablebase is actually
@@ -1676,7 +1675,9 @@ class Tablebase:
 
     def probe_wdl(self, board: chess.Board) -> int:
         """
-        Probes WDL tables for win/draw/loss-information.
+        Probes WDL tables for win/draw/loss information under the 50-move rule,
+        assuming the position has been reached directly after a capture or
+        pawn move.
 
         Probing is thread-safe when done with different *board* objects and
         if *board* objects are not modified during probing.
@@ -1837,12 +1838,18 @@ class Tablebase:
 
     def probe_dtz(self, board: chess.Board) -> int:
         """
-        Probes DTZ tables for distance to zero information.
+        Probes DTZ tables for
+        `DTZ50'' information with rounding <https://syzygy-tables.info/metrics#dtz>`_.
 
-        Both DTZ and WDL tables are required in order to probe for DTZ.
+        Minmaxing the DTZ50'' values guarantees winning a won position
+        (and drawing a drawn position), because it makes progress keeping the
+        win in hand.
+        However, the lines are not always the most straightforward ways to win.
+        Engines like Stockfish calculate themselves, checking with DTZ, but
+        only play according to DTZ if they can not manage on their own.
 
         Returns a positive value if the side to move is winning, ``0`` if the
-        position is a draw and a negative value if the side to move is losing.
+        position is a draw, and a negative value if the side to move is losing.
         More precisely:
 
         +-----+------------------+--------------------------------------------+
@@ -1872,15 +1879,14 @@ class Tablebase:
         The return value can be off by one: a return value -n can mean a
         losing zeroing move in in n + 1 plies and a return value +n can mean a
         winning zeroing move in n + 1 plies.
-        This is guaranteed not to happen for positions exactly on the edge of
-        the 50-move rule, so that (with some care) this never impacts the
-        result of practical play.
+        This implies some primary tablebase lines may waste up to 1 ply.
+        Rounding is never used for endgame phases where it would change the
+        game theoretical outcome.
 
-        Minmaxing the DTZ values guarantees winning a won position (and drawing
-        a drawn position), because it makes progress keeping the win in hand.
-        However the lines are not always the most straightforward ways to win.
-        Engines like Stockfish calculate themselves, checking with DTZ, but
-        only play according to DTZ if they can not manage on their own.
+        This means users need to be careful in positions that are nearly drawn
+        under the 50-move rule! Carelessly wasting 1 more ply by not following
+        the tablebase recommendation, for a total of 2 wasted plies, may
+        change the outcome of the game.
 
         >>> import chess
         >>> import chess.syzygy
@@ -1894,6 +1900,8 @@ class Tablebase:
         Probing is thread-safe when done with different *board* objects and
         if *board* objects are not modified during probing.
 
+        Both DTZ and WDL tables are required in order to probe for DTZ.
+
         :raises: :exc:`KeyError` (or specifically
             :exc:`chess.syzygy.MissingTableError`) if the position could not
             be found in the tablebase. Use
@@ -1961,7 +1969,7 @@ class Tablebase:
 
         self.lru.clear()
 
-    def __enter__(self) -> "Tablebase":
+    def __enter__(self) -> Tablebase:
         return self
 
     def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], traceback: Optional[TracebackType]) -> None:
