@@ -1,5 +1,5 @@
 # This file is part of the python-chess library.
-# Copyright (C) 2016-2020 Niklas Fiekas <niklas.fiekas@backscattering.de>
+# Copyright (C) 2016-2021 Niklas Fiekas <niklas.fiekas@backscattering.de>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -14,6 +14,8 @@
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
 
+from __future__ import annotations
+
 import chess
 import copy
 import itertools
@@ -318,12 +320,13 @@ class AtomicBoard(chess.Board):
             status &= ~chess.STATUS_NO_BLACK_KING
         if chess.popcount(self.checkers_mask()) <= 14:
             status &= ~chess.STATUS_TOO_MANY_CHECKERS
+        status &= ~chess.STATUS_IMPOSSIBLE_CHECK
         return status
 
 
 class KingOfTheHillBoard(chess.Board):
 
-    aliases = ["King of the Hill", "KOTH"]
+    aliases = ["King of the Hill", "KOTH", "kingOfTheHill"]
     uci_variant = "kingofthehill"
     xboard_variant = "kingofthehill"  # Unofficial
 
@@ -395,7 +398,8 @@ class RacingKingsBoard(chess.Board):
 
     def is_variant_win(self) -> bool:
         in_goal = self.kings & chess.BB_RANK_8
-        return (self.is_variant_end() and
+        return (
+            self.is_variant_end() and
             bool(in_goal & self.occupied_co[self.turn]) and
             not in_goal & self.occupied_co[not self.turn])
 
@@ -453,8 +457,210 @@ class HordeBoard(chess.Board):
         return bool(self.occupied) and not self.occupied_co[not self.turn]
 
     def has_insufficient_material(self, color: chess.Color) -> bool:
-        # TODO: Could detect some cases where the Horde can no longer mate.
-        return False
+        # The side with the king can always win by capturing the Horde.
+        if color == chess.BLACK:
+            return False
+
+        # See https://github.com/stevepapazis/horde-insufficient-material-tests
+        # for how the following has been derived.
+
+        white = self.occupied_co[chess.WHITE]
+        queens = chess.popcount(white & self.queens)
+        pawns = chess.popcount(white & self.pawns)
+        rooks = chess.popcount(white & self.rooks)
+        bishops = chess.popcount(white & self.bishops)
+        knights = chess.popcount(white & self.knights)
+
+        # Two same color bishops suffice to cover all the light and dark
+        # squares around the enemy king.
+        horde_darkb = chess.popcount(chess.BB_DARK_SQUARES & white & self.bishops)
+        horde_lightb = chess.popcount(chess.BB_LIGHT_SQUARES & white & self.bishops)
+        horde_bishop_co = chess.WHITE if horde_lightb >= 1 else chess.BLACK
+        horde_num = (
+            pawns + knights + rooks + queens +
+            (horde_darkb if horde_darkb <= 2 else 2) +
+            (horde_lightb if horde_lightb <= 2 else 2)
+        )
+
+        pieces = self.occupied_co[chess.BLACK]
+        pieces_pawns = chess.popcount(pieces & self.pawns)
+        pieces_bishops = chess.popcount(pieces & self.bishops)
+        pieces_knights = chess.popcount(pieces & self.knights)
+        pieces_rooks = chess.popcount(pieces & self.rooks)
+        pieces_queens = chess.popcount(pieces & self.queens)
+        pieces_darkb = chess.popcount(chess.BB_DARK_SQUARES & pieces & self.bishops)
+        pieces_lightb = chess.popcount(chess.BB_LIGHT_SQUARES & pieces & self.bishops)
+        pieces_num = chess.popcount(pieces)
+
+        def pieces_oppositeb_of(square_color: chess.Color) -> int:
+            return pieces_darkb if square_color == chess.WHITE else pieces_lightb
+
+        def pieces_sameb_as(square_color: chess.Color) -> int:
+            return pieces_lightb if square_color == chess.WHITE else pieces_darkb
+
+        def pieces_of_type_not(piece: int) -> int:
+            return pieces_num - piece
+
+        def has_bishop_pair(side: chess.Color) -> bool:
+            return (horde_lightb >= 1 and horde_darkb >= 1) if side == chess.WHITE else (pieces_lightb >= 1 and pieces_darkb >= 1)
+
+        if horde_num == 0:
+            return True
+        if horde_num >= 4:
+            # Four or more white pieces can always deliver mate.
+            return False
+        if (pawns >= 1 or queens >= 1) and horde_num >= 2:
+            # Pawns/queens are never insufficient material when paired with any other
+            # piece (a pawn promotes to a queen and delivers mate).
+            return False
+        if rooks >= 1 and horde_num >= 2:
+            # A rook is insufficient material only when it is paired with a bishop
+            # against a lone king. The horde can mate in any other case.
+            # A rook on A1 and a bishop on C3 mate a king on B1 when there is a
+            # friendly pawn/opposite-color-bishop/rook/queen on C2.
+            # A rook on B8 and a bishop C3 mate a king on A1 when there is a friendly
+            # knight on A2.
+            if not (horde_num == 2 and rooks == 1 and bishops == 1 and pieces_of_type_not(pieces_sameb_as(horde_bishop_co)) == 1):
+                return False
+
+        if horde_num == 1:
+            if pieces_num == 1:
+                # A lone piece cannot mate a lone king.
+                return True
+            elif queens == 1:
+                # The horde has a lone queen.
+                # A lone queen mates a king on A1 bounded by:
+                # - a pawn/rook on A2
+                # - two same color bishops on A2, B1
+                # We ignore every other mating case, since it can be reduced to
+                # the two previous cases (e.g. a black pawn on A2 and a black
+                # bishop on B1).
+                return not (
+                    pieces_pawns >= 1 or
+                    pieces_rooks >= 1 or
+                    pieces_lightb >= 2 or
+                    pieces_darkb >= 2
+                )
+            elif pawns == 1:
+                # Promote the pawn to a queen or a knight and check whether
+                # white can mate.
+                pawn_square = chess.SquareSet(self.pawns & white).pop()
+                promote_to_queen = self.copy(stack=False)
+                promote_to_queen.set_piece_at(pawn_square, chess.Piece(chess.QUEEN, chess.WHITE))
+                promote_to_knight = self.copy(stack=False)
+                promote_to_knight.set_piece_at(pawn_square, chess.Piece(chess.KNIGHT, chess.WHITE))
+                return promote_to_queen.has_insufficient_material(chess.WHITE) and promote_to_knight.has_insufficient_material(chess.WHITE)
+            elif rooks == 1:
+                # A lone rook mates a king on A8 bounded by a pawn/rook on A7 and a
+                # pawn/knight on B7. We ignore every other case, since it can be
+                # reduced to the two previous cases.
+                # (e.g. three pawns on A7, B7, C7)
+                return not (
+                    pieces_pawns >= 2 or
+                    (pieces_rooks >= 1 and pieces_pawns >= 1) or
+                    (pieces_rooks >= 1 and pieces_knights >= 1) or
+                    (pieces_pawns >= 1 and pieces_knights >= 1)
+                )
+            elif bishops == 1:
+                # The horde has a lone bishop.
+                return not (
+                    # The king can be mated on A1 if there is a pawn/opposite-color-bishop
+                    # on A2 and an opposite-color-bishop on B1.
+                    # If black has two or more pawns, white gets the benefit of the doubt;
+                    # there is an outside chance that white promotes its pawns to
+                    # opposite-color-bishops and selfmates theirself.
+                    # Every other case that the king is mated by the bishop requires that
+                    # black has two pawns or two opposite-color-bishop or a pawn and an
+                    # opposite-color-bishop.
+                    # For example a king on A3 can be mated if there is
+                    # a pawn/opposite-color-bishop on A4, a pawn/opposite-color-bishop on
+                    # B3, a pawn/bishop/rook/queen on A2 and any other piece on B2.
+                    pieces_oppositeb_of(horde_bishop_co) >= 2 or
+                    (pieces_oppositeb_of(horde_bishop_co) >= 1 and pieces_pawns >= 1) or
+                    pieces_pawns >= 2
+                )
+            elif knights == 1:
+                # The horde has a lone knight.
+                return not (
+                    # The king on A1 can be smother mated by a knight on C2 if there is
+                    # a pawn/knight/bishop on B2, a knight/rook on B1 and any other piece
+                    # on A2.
+                    # Moreover, when black has four or more pieces and two of them are
+                    # pawns, black can promote their pawns and selfmate theirself.
+                    pieces_num >= 4 and (
+                        pieces_knights >= 2 or pieces_pawns >= 2 or
+                        (pieces_rooks >= 1 and pieces_knights >= 1) or
+                        (pieces_rooks >= 1 and pieces_bishops >= 1) or
+                        (pieces_knights >= 1 and pieces_bishops >= 1) or
+                        (pieces_rooks >= 1 and pieces_pawns >= 1) or
+                        (pieces_knights >= 1 and pieces_pawns >= 1) or
+                        (pieces_bishops >= 1 and pieces_pawns >= 1) or
+                        (has_bishop_pair(chess.BLACK) and pieces_pawns >= 1)
+                    ) and
+                    (pieces_of_type_not(pieces_darkb) >= 3 if pieces_darkb >= 2 else True) and
+                    (pieces_of_type_not(pieces_lightb) >= 3 if pieces_lightb >= 2 else True)
+                )
+        elif horde_num == 2:  # By this point, we only need to deal with white's minor pieces.
+            if pieces_num == 1:
+                # Two minor pieces cannot mate a lone king.
+                return True
+            elif knights == 2:
+                # A king on A1 is mated by two knights, if it is obstructed by a
+                # pawn/bishop/knight on B2. On the other hand, if black only has
+                # major pieces it is a draw.
+                return not (pieces_pawns + pieces_bishops + pieces_knights >= 1)
+            elif has_bishop_pair(chess.WHITE):
+                return not (
+                    # A king on A1 obstructed by a pawn/bishop on A2 is mated
+                    # by the bishop pair.
+                    pieces_pawns >= 1 or pieces_bishops >= 1 or
+                    # A pawn/bishop/knight on B4, a pawn/bishop/rook/queen on
+                    # A4 and the king on A3 enable Boden's mate by the bishop
+                    # pair. In every other case white cannot win.
+                    (pieces_knights >= 1 and pieces_rooks + pieces_queens >= 1)
+                )
+            elif bishops >= 1 and knights >= 1:
+                # The horde has a bishop and a knight.
+                return not (
+                    # A king on A1 obstructed by a pawn/opposite-color-bishop on
+                    # A2 is mated by a knight on D2 and a bishop on C3.
+                    pieces_pawns >= 1 or pieces_oppositeb_of(horde_bishop_co) >= 1 or
+                    # A king on A1 bounded by two friendly pieces on A2 and B1 is
+                    # mated when the knight moves from D4 to C2 so that both the
+                    # knight and the bishop deliver check.
+                    pieces_of_type_not(pieces_sameb_as(horde_bishop_co)) >= 3
+                )
+            else:
+                # The horde has two or more bishops on the same color.
+                # White can only win if black has enough material to obstruct
+                # the squares of the opposite color around the king.
+                return not (
+                    # A king on A1 obstructed by a pawn/opposite-bishop/knight
+                    # on A2 and a opposite-bishop/knight on B1 is mated by two
+                    # bishops on B2 and C3. This position is theoretically
+                    # achievable even when black has two pawns or when they
+                    # have a pawn and an opposite color bishop.
+                    (pieces_pawns >= 1 and pieces_oppositeb_of(horde_bishop_co) >= 1) or
+                    (pieces_pawns >= 1 and pieces_knights >= 1) or
+                    (pieces_oppositeb_of(horde_bishop_co) >= 1 and pieces_knights >= 1) or
+                    (pieces_oppositeb_of(horde_bishop_co) >= 2) or
+                    pieces_knights >= 2 or
+                    pieces_pawns >= 2
+                    # In every other case, white can only draw.
+                )
+        elif horde_num == 3:
+            # A king in the corner is mated by two knights and a bishop or three
+            # knights or the bishop pair and a knight/bishop.
+            if (knights == 2 and bishops == 1) or knights == 3 or has_bishop_pair(chess.WHITE):
+                return False
+            else:
+                # White has two same color bishops and a knight.
+                # A king on A1 is mated by a bishop on B2, a bishop on C1 and a
+                # knight on C3, as long as there is another black piece to waste
+                # a tempo.
+                return pieces_num == 1
+
+        return True
 
     def status(self) -> chess.Status:
         status = super().status()
@@ -475,20 +681,20 @@ class HordeBoard(chess.Board):
 
 ThreeCheckBoardT = TypeVar("ThreeCheckBoardT", bound="ThreeCheckBoard")
 
-class _ThreeCheckBoardState(Generic[ThreeCheckBoardT], chess._BoardState["ThreeCheckBoardT"]):
-    def __init__(self, board: "ThreeCheckBoardT") -> None:
+class _ThreeCheckBoardState(Generic[ThreeCheckBoardT], chess._BoardState[ThreeCheckBoardT]):
+    def __init__(self, board: ThreeCheckBoardT) -> None:
         super().__init__(board)
         self.remaining_checks_w = board.remaining_checks[chess.WHITE]
         self.remaining_checks_b = board.remaining_checks[chess.BLACK]
 
-    def restore(self, board: "ThreeCheckBoardT") -> None:
+    def restore(self, board: ThreeCheckBoardT) -> None:
         super().restore(board)
         board.remaining_checks[chess.WHITE] = self.remaining_checks_w
         board.remaining_checks[chess.BLACK] = self.remaining_checks_b
 
 class ThreeCheckBoard(chess.Board):
 
-    aliases = ["Three-check", "Three check", "Threecheck", "Three check chess", "3-check"]
+    aliases = ["Three-check", "Three check", "Threecheck", "Three check chess", "3-check", "3 check", "3check"]
     uci_variant = "3check"
     xboard_variant = "3check"
     starting_fen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 3+3 0 1"
@@ -564,7 +770,7 @@ class ThreeCheckBoard(chess.Board):
         self.remaining_checks[chess.WHITE] = wc
         self.remaining_checks[chess.BLACK] = bc
 
-    def epd(self, shredder: bool = False, en_passant: str = "legal", promoted: Optional[bool] = None, **operations: Union[None, str, int, float, chess.Move, Iterable[chess.Move]]) -> str:
+    def epd(self, shredder: bool = False, en_passant: chess._EnPassantSpec = "legal", promoted: Optional[bool] = None, **operations: Union[None, str, int, float, chess.Move, Iterable[chess.Move]]) -> str:
         epd = [super().epd(shredder=shredder, en_passant=en_passant, promoted=promoted),
                "{:d}+{:d}".format(max(self.remaining_checks[chess.WHITE], 0),
                                   max(self.remaining_checks[chess.BLACK], 0))]
@@ -605,16 +811,16 @@ class ThreeCheckBoard(chess.Board):
 
 CrazyhouseBoardT = TypeVar("CrazyhouseBoardT", bound="CrazyhouseBoard")
 
-class _CrazyhouseBoardState(Generic[CrazyhouseBoardT], chess._BoardState["CrazyhouseBoardT"]):
-    def __init__(self, board: "CrazyhouseBoardT") -> None:
+class _CrazyhouseBoardState(Generic[CrazyhouseBoardT], chess._BoardState[CrazyhouseBoardT]):
+    def __init__(self, board: CrazyhouseBoardT) -> None:
         super().__init__(board)
         self.pockets_w = board.pockets[chess.WHITE].copy()
         self.pockets_b = board.pockets[chess.BLACK].copy()
 
-    def restore(self, board: "CrazyhouseBoardT") -> None:
+    def restore(self, board: CrazyhouseBoardT) -> None:
         super().restore(board)
-        board.pockets[chess.WHITE] = self.pockets_w.copy()
-        board.pockets[chess.BLACK] = self.pockets_b.copy()
+        board.pockets[chess.WHITE] = self.pockets_w
+        board.pockets[chess.BLACK] = self.pockets_b
 
 CrazyhousePocketT = TypeVar("CrazyhousePocketT", bound="CrazyhousePocket")
 
@@ -697,10 +903,8 @@ class CrazyhouseBoard(chess.Board):
         else:
             self.pockets[self.turn].add(piece_type)
 
-    def can_claim_fifty_moves(self) -> bool:
-        return False
-
-    def is_seventyfive_moves(self) -> bool:
+    def _is_halfmoves(self, n: int) -> bool:
+        # No draw by 50-move rule or 75-move rule.
         return False
 
     def is_irreversible(self, move: chess.Move) -> bool:
@@ -726,6 +930,15 @@ class CrazyhouseBoard(chess.Board):
             return chess.BB_EMPTY
 
     def legal_drop_squares(self) -> chess.SquareSet:
+        """
+        Gets the squares where the side to move could legally drop a piece.
+        Does *not* check whether they actually have a suitable piece in their
+        pocket.
+
+        It is legal to drop a checkmate.
+
+        Returns a :class:`set of squares <chess.SquareSet>`.
+        """
         return chess.SquareSet(self.legal_drop_squares_mask())
 
     def is_pseudo_legal(self, move: chess.Move) -> bool:
@@ -760,7 +973,7 @@ class CrazyhouseBoard(chess.Board):
 
     def parse_san(self, san: str) -> chess.Move:
         if "@" in san:
-            uci = san.rstrip("+# ")
+            uci = san.rstrip("+#")
             if uci[0] == "@":
                 uci = "P" + uci
             move = chess.Move.from_uci(uci)
@@ -771,8 +984,8 @@ class CrazyhouseBoard(chess.Board):
             return super().parse_san(san)
 
     def has_insufficient_material(self, color: chess.Color) -> bool:
-        # In practise no material can leave the game, but this is easy to
-        # implement anyway. Note that bishops can be captured and put onto
+        # In practice, no material can leave the game, but this is easy to
+        # implement, anyway. Note that bishops can be captured and put onto
         # a different color complex.
         return (
             chess.popcount(self.occupied) + sum(len(pocket) for pocket in self.pockets) <= 3 and
@@ -813,7 +1026,7 @@ class CrazyhouseBoard(chess.Board):
             promoted = True
         return super().board_fen(promoted=promoted)
 
-    def epd(self, shredder: bool = False, en_passant: str = "legal", promoted: Optional[bool] = None, **operations: Union[None, str, int, float, chess.Move, Iterable[chess.Move]]) -> str:
+    def epd(self, shredder: bool = False, en_passant: chess._EnPassantSpec = "legal", promoted: Optional[bool] = None, **operations: Union[None, str, int, float, chess.Move, Iterable[chess.Move]]) -> str:
         epd = super().epd(shredder=shredder, en_passant=en_passant, promoted=promoted)
         board_part, info_part = epd.split(" ", 1)
         return f"{board_part}[{str(self.pockets[chess.WHITE]).upper()}{self.pockets[chess.BLACK]}] {info_part}"
@@ -857,7 +1070,10 @@ VARIANTS: List[Type[chess.Board]] = [
 
 
 def find_variant(name: str) -> Type[chess.Board]:
-    """Looks for a variant board class by variant name."""
+    """
+    Looks for a variant board class by variant name. Supports many common
+    aliases.
+    """
     for variant in VARIANTS:
         if any(alias.lower() == name.lower() for alias in variant.aliases):
             return variant
